#pragma once

#include <samurai/field_expression.hpp>

namespace samurai
{

    struct system_expression_tag
    {
    };

    template <class E>
    struct is_system_expression : std::is_same<xt::xexpression_tag_t<E>, system_expression_tag>
    {
    };

    template <class... E>
    struct system_comparable : xtl::conjunction<is_system_expression<E>...>
    {
    };

    template <class D>
    class system_expression : public xt::xexpression<D>
    {
      public:

        using expression_tag = system_expression_tag;
    };

    template <class T1, class... T>
    struct get_temporary_type
    {
        using type = std::conditional_t<is_system_expression<T1>::value, xt::temporary_type_t<T1>, typename get_temporary_type<T...>::type>;
    };

    template <class T1>
    struct get_temporary_type<T1>
    {
        using type = std::conditional_t<is_system_expression<T1>::value, xt::temporary_type_t<T1>, void>;
    };

    template <class... T>
    using get_temporary_type_t = typename get_temporary_type<T...>::type;

    template <class F, class... CT>
    class system_function : public system_expression<system_function<F, CT...>>
    {
      public:

        using self_type    = system_function<F, CT...>;
        using functor_type = std::remove_reference_t<F>;

        using temporary_type = get_temporary_type_t<CT...>;
        using expression_tag = system_expression_tag;

        template <class Func, class... CTA, class U = std::enable_if<!std::is_base_of<Func, self_type>::value>>
        system_function(Func&& f, CTA&&... e) noexcept;

        const auto& arguments() const
        {
            return m_e;
        }

        const auto& function() const
        {
            return m_f;
        }

      private:

        std::tuple<CT...> m_e;
        functor_type m_f;
    };

    template <class F, class... CT>
    template <class Func, class... CTA, class>
    inline system_function<F, CT...>::system_function(Func&& f, CTA&&... e) noexcept
        : m_e(std::forward<CTA>(e)...)
        , m_f(std::forward<Func>(f))
    {
    }
}

namespace xt
{
    namespace detail
    {
        template <class F, class... E>
        struct select_xfunction_expression<samurai::system_expression_tag, F, E...>
        {
            using type = samurai::system_function<F, E...>;
        };
    }

    template <class T>
    struct temporary_type_from_tag<samurai::system_expression_tag, T>
    {
        using type = double;
    };
} // namespace xt
