#pragma once

#include "tuple_expression.hpp"
#include "utils.hpp"
#include <tuple>
#include <xtensor/xscalar.hpp>

#include <type_traits>

using namespace samurai::literals;

namespace samurai
{
    template <class T>
    auto get_value(T&& t)
    {
        return std::forward<T>(t);
    }

    double get_value(double t)
    {
        return t;
    }

    template <class T>
    typename std::enable_if<xt::is_xscalar<T>::value, typename std::decay_t<T>::value_type>::type get_value(const T& t)
    {
        return t();
    }

    template <std::size_t i, class T, class U = std::conditional_t<std::is_arithmetic_v<T>, T, const T&>>
    typename std::enable_if<!is_tuple_expression<T>::value, U>::type get_tuple_entry(const T& t)
    {
        // std::cout << "const Other " << i << std::endl;
        return t;
    }

    template <std::size_t i, class T, class U = std::conditional_t<std::is_arithmetic_v<T>, T&, T&>>
    typename std::enable_if<!is_tuple_expression<T>::value, U>::type get_tuple_entry(T& t)
    {
        // std::cout << "const Other " << i << std::endl;
        return t;
    }

    template <std::size_t I, class... Types>
    decltype(auto) get_tuple_entry(std::tuple<Types...>& t)
    {
        // std::cout << "tuple " << I << std::endl;
        return get_tuple_entry<I>(std::get<I>(t));
    }

    // template <std::size_t I, class... Types>
    // auto& get_tuple_entry(std::tuple<Types...>& t)
    // {
    //     // std::cout << "tuple " << I << std::endl;
    //     return get_tuple_entry<I>(std::get<I>(t));
    // }

    template <std::size_t I, class... Types>
    decltype(auto) get_tuple_entry(const std::tuple<Types...>& t)
    {
        // std::cout << "const tuple " << I << std::endl;
        return get_tuple_entry<I>(std::get<I>(t));
    }

    template <std::size_t i, class F, class Tuple, std::size_t... I>
    auto myapply_tuple_impl(F&&, Tuple&& t, std::index_sequence<I...>)
    {
        return xt::detail::make_xfunction<std::decay_t<F>>(get_tuple_entry<i>(std::get<I>(t))...);
    }

    template <std::size_t i, class F, class Tuple>
    auto myapply_tuple(F&& f, Tuple&& t)
    {
        return myapply_tuple_impl<i>(std::forward<F>(f),
                                     std::forward<Tuple>(t),
                                     std::make_index_sequence<std::tuple_size_v<std::decay_t<Tuple>>>());
    }

    template <std::size_t i, class F, class... CT>
    auto get_tuple_entry(const tuple_function<F, CT...>& e)
    {
        // std::cout << "const tuple_function " << i << std::endl;
        return myapply_tuple<i>(e.function(), e.arguments());
    }

    template <std::size_t i, class E>
    decltype(auto) get_tuple_entry(tuple_expression<E>& e)
    {
        // std::cout << "tuple_expression " << i << std::endl;
        return get_tuple_entry<i>(e.derived_cast().arguments());
    }

    template <std::size_t i, class E>
    decltype(auto) get_tuple_entry(const tuple_expression<E>& e)
    {
        // std::cout << "const tuple_expression " << i << std::endl;
        return get_tuple_entry<i>(e.derived_cast().arguments());
    }

    namespace detail
    {
        template <class Dest, class Src>
        void final_assign(Dest& dst, const Src& src)
        {
            dst = src;
        }

        template <class Src>
        void final_assign(double& dst, const Src& src)
        {
            dst = src[0];
        }

        void final_assign(double& dst, double src)
        {
            dst = src;
        }
    }

    template <std::size_t I, class System, class Expression>
    void assign_tuple_on_entry(System& sys, Expression& e)
    {
        detail::final_assign(get_tuple_entry<I>(sys), get_tuple_entry<I>(e));
        // get_tuple_entry<I>(sys) = get_tuple_entry<I>(e)[0];
    }

    template <class System, class Expression, std::size_t... I>
    void assign_tuple_impl(System& sys, Expression& e, std::index_sequence<I...>)
    {
        (assign_tuple_on_entry<I>(sys, e), ...);
    }

    template <class System, class Expression>
    void assign_tuple(System& sys, Expression& e)
    {
        assign_tuple_impl(sys, e, std::make_index_sequence<System::size>());
    }

    namespace detail
    {

        template <class T, bool>
        struct get_temporary
        {
            using type = T;
        };

        // template <class CT, class... S>
        // struct get_temporary<xt::xview<CT, S...>>
        // {
        //     using type = typename xt::xview<CT, S...>::temporary_type;
        // };

        template <class T>
        struct get_temporary<T, true>
        {
            using type = typename T::temporary_type;
        };

        template <class T>
        using get_temporary_t = typename get_temporary<T, xt::is_xexpression<std::decay_t<T>>::value>::type;

    }

    template <class... T>
    class Tuple_container_impl : public tuple_expression<Tuple_container_impl<T...>>
    {
      public:

        using tuple_t                     = std::tuple<T...>;
        using concrete_t                  = tuple_t;
        using temporary_type              = Tuple_container_impl<detail::get_temporary_t<std::decay_t<T>>...>;
        static constexpr std::size_t size = sizeof...(T);

        Tuple_container_impl() = default;

        template <class... CTA>
        Tuple_container_impl(CTA&&... t)
            : m_elements(std::forward<CTA>(t)...)
        {
        }

        template <class F, class... CT>
        Tuple_container_impl(tuple_function<F, CT...>& e)
        {
            temporary_type out;

            assign(out, e.derived_cast());
            std::swap(*this, out);
        }

        template <class F, class... CT>
        Tuple_container_impl(tuple_function<F, CT...>&& e)
        {
            temporary_type out;

            assign(out, e.derived_cast());
            std::swap(*this, out);
        }

        template <class F, class... CT>
        Tuple_container_impl(const tuple_function<F, CT...>& e)
        {
            temporary_type out;

            assign(out, e.derived_cast());
            std::swap(*this, out);
        }

        template <class E>
        void operator=(const tuple_expression<E>& e)
        {
            assign_tuple(*this, e.derived_cast());
        }

        template <class E>
        void operator+=(const tuple_expression<E>& e)
        {
            auto expr = *this + e;
            assign_tuple(*this, expr.derived_cast());
        }

        template <class E>
        void operator-=(const tuple_expression<E>& e)
        {
            auto expr = *this - e;
            assign_tuple(*this, expr.derived_cast());
        }

        template <std::size_t I>
        auto& operator[](index<I>)
        {
            static_assert(I < size);
            return std::get<I>(m_elements);
        }

        template <std::size_t I>
        const auto& operator[](index<I>) const
        {
            static_assert(I < size);
            return std::get<I>(m_elements);
        }

        // template <std::size_t I>
        // const double operator[](index<I>) const
        // {
        //     static_assert(I < size);
        //     return std::get<I>(m_elements);
        // }

        const auto& arguments() const
        {
            return m_elements;
        }

        auto& arguments()
        {
            return m_elements;
        }

        void to_stream(std::ostream& os) const
        {
            os << "Tuple_container" << std::endl;
            std::apply(
                [&os](auto&... e)
                {
                    (os << ... << fmt::format("{} ", e)) << std::endl;
                },
                m_elements);
        }

      private:

        concrete_t m_elements;
    };

    template <bool by_ref, class T>
    struct unwrap_refwrapper
    {
        using type = T;
    };

    template <>
    struct unwrap_refwrapper<true, double>
    {
        using type = double&;
    };

    template <>
    struct unwrap_refwrapper<false, double>
    {
        using type = double;
    };

    template <bool by_ref, class T>
    struct unwrap_refwrapper<by_ref, std::reference_wrapper<T>>
    {
        using type = T&;
    };

    template <bool by_ref, class T>
    using unwrap_decay_t = typename unwrap_refwrapper<by_ref, typename std::decay<T>::type>::type;

    template <bool by_ref, class... T>
    auto Tuple_container(T&&... t)
    {
        // static_assert(std::is_same_v<std::conjunction<T...>, bool>);

        // static_assert(std::is_same_v<Tuple_container_impl<unwrap_decay_t<T>...>, bool>);
        return Tuple_container_impl<unwrap_decay_t<by_ref, T>...>(std::forward<T>(t)...);
    }

    template <class... T>
    inline std::ostream& operator<<(std::ostream& out, const Tuple_container_impl<T...>& t)
    {
        t.to_stream(out);
        return out;
    }
}