#pragma once

#include <samurai/field_expression.hpp>

namespace samurai
{

    struct tuple_expression_tag
    {
    };

    template <class E>
    struct is_tuple_expression : std::is_same<xt::xexpression_tag_t<E>, tuple_expression_tag>
    {
    };

    template <class... E>
    struct tuple_comparable : xtl::conjunction<is_tuple_expression<E>...>
    {
    };

    template <class D>
    class tuple_expression : public xt::xexpression<D>
    {
      public:

        using expression_tag = tuple_expression_tag;
    };

    template <class T1, class... T>
    struct get_tuple_temporary_type
    {
        using type = std::conditional_t<is_tuple_expression<T1>::value, xt::temporary_type_t<T1>, typename get_tuple_temporary_type<T...>::type>;
    };

    template <class T1>
    struct get_tuple_temporary_type<T1>
    {
        using type = std::conditional_t<is_tuple_expression<T1>::value, xt::temporary_type_t<T1>, void>;
    };

    template <class... T>
    using get_tuple_temporary_type_t = typename get_tuple_temporary_type<T...>::type;

    template <class F, class... CT>
    class tuple_function : public tuple_expression<tuple_function<F, CT...>>
    {
      public:

        using self_type    = tuple_function<F, CT...>;
        using functor_type = std::remove_reference_t<F>;

        using temporary_type = get_tuple_temporary_type_t<CT...>;
        using expression_tag = tuple_expression_tag;

        template <class Func, class... CTA, class U = std::enable_if<!std::is_base_of<Func, self_type>::value>>
        tuple_function(Func&& f, CTA&&... e) noexcept;

        const auto& arguments() const
        {
            return m_e;
        }

        const auto& function() const
        {
            return m_f;
        }

      private:

        std::tuple<CT...> m_e;
        functor_type m_f;
    };

    template <class F, class... CT>
    template <class Func, class... CTA, class>
    inline tuple_function<F, CT...>::tuple_function(Func&& f, CTA&&... e) noexcept
        : m_e(std::forward<CTA>(e)...)
        , m_f(std::forward<Func>(f))
    {
    }
}

namespace xt
{
    namespace detail
    {
        template <class F, class... E>
        struct select_xfunction_expression<samurai::tuple_expression_tag, F, E...>
        {
            using type = samurai::tuple_function<F, E...>;
        };
    }

    template <class T>
    struct temporary_type_from_tag<samurai::tuple_expression_tag, T>
    {
        using type = double;
    };
} // namespace xt
