#pragma once

#include <tuple>
#include <utility>

namespace samurai
{
    namespace detail
    {
        template <class Tuple, class Func, std::size_t... I>
        void for_each_impl(Tuple&& t, Func&& f, std::index_sequence<I...>)
        {
            (f(std::get<I>(t)), ...);
        }

    }

    template <class Tuple, class Func>
    void for_each(Tuple&& t, Func&& f)
    {
        detail::for_each_impl(std::forward<Tuple>(t),
                              std::forward<Func>(f),
                              std::make_index_sequence<std::tuple_size_v<std::decay_t<Tuple>>>());
    }

    template <std::size_t I>
    struct index
    {
    };

    // Take from boost::hana

    //////////////////////////////////////////////////////////////////////////
    // User-defined literal
    //////////////////////////////////////////////////////////////////////////
    namespace ic_detail
    {

        constexpr int to_int(char c)
        {
            int result = 0;

            if (c >= 'A' && c <= 'F')
            {
                result = static_cast<int>(c) - static_cast<int>('A') + 10;
            }
            else if (c >= 'a' && c <= 'f')
            {
                result = static_cast<int>(c) - static_cast<int>('a') + 10;
            }
            else
            {
                result = static_cast<int>(c) - static_cast<int>('0');
            }

            return result;
        }

        template <std::size_t N>
        constexpr std::size_t parse(const char (&arr)[N])
        {
            long long base     = 10;
            std::size_t offset = 0;

            if (N > 2)
            {
                bool starts_with_zero = arr[0] == '0';
                bool is_hex           = starts_with_zero && (arr[1] == 'x' || arr[1] == 'X');
                bool is_binary        = starts_with_zero && arr[1] == 'b';

                if (is_hex)
                {
                    // 0xDEADBEEF (hexadecimal)
                    base   = 16;
                    offset = 2;
                }
                else if (is_binary)
                {
                    // 0b101011101 (binary)
                    base   = 2;
                    offset = 2;
                }
                else if (starts_with_zero)
                {
                    // 012345 (octal)
                    base   = 8;
                    offset = 1;
                }
            }

            long long number     = 0;
            long long multiplier = 1;

            for (std::size_t i = 0; i < N - offset; ++i)
            {
                char c = arr[N - 1 - i];
                if (c != '\'')
                { // skip digit separators
                    number += to_int(c) * multiplier;
                    multiplier *= base;
                }
            }

            return static_cast<std::size_t>(number);
        }
    }

    namespace literals
    {
        template <char... c>
        constexpr auto operator"" _c()
        {
            return index<ic_detail::parse<sizeof...(c)>({c...})>{};
        }
    }

}